#
# crnsimulator library: simulate chemical reaction networks using ODEs
#
# Written by Stefan Badelt (badelt@caltech.edu).
#
# Use at your own risk.
#
#
__version__ = "v0.7.1"

from crnsimulator.crn_parser import parse_crn_string, parse_crn_file
from crnsimulator.reactiongraph import ReactionGraph
from crnsimulator.solver import writeODElib, get_integrator
from crnsimulator.odelib_template import ode_plotter



#
# CRN-to-MultiDiGraph-to-ODE translation utilities.
#
# Written by Stefan Badelt (badelt@caltech.edu).
#
# Use at your own risk.
#

import types
import importlib.machinery

import crnsimulator.odelib_template

def get_integrator(filename, function = 'integrate'):
    """ Wrapper for the jit import of a function from a python script.
    
    Note: The intended usage is to import the integrate function from the
    executable produced by crnsimulator. However, it can be used to import
    arbitrary functions from arbitrary python scripts.

    Args:
        filename (str): The autogenerated python executable containing the odesystem. 
        function (str, optional): The name of the function to be imported. Defaults
            to 'integrate', which is the main function of the file, but it can also 
            be used to import low-level functions such as 'odesystem' or 'ode_plotter'.

    Returns:
        A jit import of the requested function.
    """
    try:
        loader = importlib.machinery.SourceFileLoader('my_loader', filename)
        mod = types.ModuleType(loader.name)
        loader.exec_module(mod)
    except FileNotFoundError as err:
        print(Warning('DEPRECATED - crnsimulator: Please note that the crnsimulator.solver.get_integrator function interface changed.'))
        raise err

    return getattr(mod, function)

def writeODElib(svars, odeM, jacobian=None, rdict=None, concvect=None,
                odename='odesystem', filename='./odesystem', template=None):
    """ Write an ODE system into an executable python script.

    Args:
      svars <list[str]>: Sorted list of variables. The sorting defines the order
        for specifying concentrations.
      odeM <sympy.Matrix()>: A matrix that contains the ODE system.
      jacobian <optional: sympy.Matrix()> : The jacobi Matrix corresponding to
        odeM.
      rdict <optional: dict()>: If your odeM contains rates in form of variable
        names, then you need to supply this dictionary mapping names to float values.
      concvect <optional: list(): Specify default initial species concentrations
        in the order defined by svars.
      odename <optional: str>: Name of your ODE function (no special characters!)
      filename <optional: str>: Specify the name of the ODE library.
      template <optional: str>: Specify an alternative template library file.

    Returns:
      filename<str>, odename<str>

    """
    if not template:
        template = crnsimulator.odelib_template.__file__[:]
        if template[-1] == 'c':
            template = template[:-1]

    svars = list(map(str, svars))

    odetemp = ''
    with open(template, 'r') as tfile:
        odetemp = tfile.read()

    # REPLACE NAMES
    odetemp = odetemp.replace("#<&>ODENAME<&>#", odename)
    odetemp = odetemp.replace("#<&>FILENAME<&>#", filename)

    # DEFAULT RATES
    ratestring = ',\n'.join(
        "  '{}' : {}".format(k, rdict[k]) for k in sorted(rdict.keys()))
    odetemp = odetemp.replace("#<&>RATES<&>#", ratestring)

    # ODEINT FUNCTION
    functionstring = "def {}(p0, t0, r):\n".format(odename)
    # Initialize arguments
    functionstring += "    {} = p0\n".format(', '.join(svars))
    functionstring += "    if not r : r = rates\n\n"
    for k in sorted(rdict.keys()):
        functionstring += "    {} = r['{}']\n".format(k, k)
    functionstring += "\n"
    # Write the ODEs
    for i in range(len(svars)):
        functionstring += "    d{}dt = {}\n".format(svars[i], odeM[i])
    # return
    if len(svars) == 1:
        functionstring += "    return np.array({})".format(
            ', '.join(['d' + x + 'dt' for x in svars]))
    else:
        functionstring += "    return np.array([{}])".format(
            ', '.join(['d' + x + 'dt' for x in svars]))
    odetemp = odetemp.replace("#<&>ODECALL<&>#", functionstring)

    if jacobian:
        # JACOBIAN FUNCTION
        jacobianstring = "def {}(p0, t0, r):\n".format('jacobian')
        # Initialize arguments
        jacobianstring += "    {} = p0\n".format(', '.join(svars))
        jacobianstring += "    if not r : r = rates\n\n"
        for k in sorted(rdict.keys()):
            jacobianstring += "    {} = r['{}']\n".format(k, k)
        jacobianstring += "\n"

        # Write the jacobian
        jacobianstring += "    J = [[[] for i in range(len(p0))] " + \
            "for j in range(len(p0))]\n"

        vl = len(svars)
        i, j = 0, 0
        for row in jacobian:
            jacobianstring += "    J[{}][{}] = {}\n".format(i, j, row)
            if j < vl - 1:
                j += 1
            else:
                i += 1
                j = 0

        if len(svars) == 1:
            jacobianstring += "    J = J[0]\n"

        # return
        jacobianstring += "    return J"
        odetemp = odetemp.replace("#<&>JACOBIAN<&>#", jacobianstring)
        odetemp = odetemp.replace("#<&>JCALL<&>#", 'Dfun = jacobian')

    # SORTED VARIABLE NAMES in integrate()
    svarstring = 'svars = ' + '[{}]'.format(
        ', '.join([str('"' + x + '"') for x in svars]))
    odetemp = odetemp.replace("#<&>SORTEDVARS<&>#", svarstring)

    # Default concentrations in integrate()
    concstring = ''

    if concvect:
        for e, c in enumerate(concvect):
            if c:
                concstring += "p0[{}] = {}\n    ".format(e, c)
    odetemp = odetemp.replace("#<&>DEFAULTCONCENTRATIONS<&>#", concstring)

    if filename[-3:] != '.py':
        filename += '.py'
    with open(filename, 'w') as ofile:
        ofile.write(odetemp)

    return filename, odename



#
# CRN-to-MultiDiGraph-to-ODE translation utilities.
#
# Written by Stefan Badelt (badelt@caltech.edu).
#
# Use at your own risk.
#
#

import networkx as nx
from typing import Dict, List, Tuple, Sequence, TypeVar, Union
from sympy import sympify, Matrix, Symbol

from crnsimulator.solver import writeODElib

# Typehintsh
SPE = List[str]
RXN = Tuple[SPE, SPE, str]
CRN = List[RXN]
sM = TypeVar('sympy.Matrix')


class CRNSimulatorError(Exception):
    def __init__(self, message):
        super(CRNSimulatorError, self).__init__(message)


class ReactionNode(object):
    """A Reaction-Node in the ReactionGraph class. """
    rid = 0
    def __init__(self, prefix: str = 'RXN:'):
        self.name = prefix + str(ReactionNode.rid)
        ReactionNode.rid += 1

class ReactionGraph(object):
    """Basic Reaction Graph Object. """

    def __init__(self, crn: CRN = None):
        self._RG = nx.MultiDiGraph()
        if crn: self.add_reactions(crn)

    @property
    def reactions(self):
        return [n for n in self._RG.nodes() if isinstance(n, ReactionNode)]

    @property
    def species(self):
        return [n for n in self._RG.nodes() if not isinstance(n, ReactionNode)]

    @property
    def reactants(self):
        return [n for n in self._RG.nodes() if not isinstance(n, ReactionNode)
                and self._RG.out_edges(n)]

    @property
    def products(self):
        return [n for n in self._RG.nodes() if not isinstance(n, ReactionNode)
                and self._RG.in_edges(n)]

    def write_ODE_lib(self, 
            sorted_vars: List[str] = None, 
            concvect: List[float] = None, 
            jacobian: bool = False, 
            rate_dict: bool = False,
            odename: str = 'odesystem', 
            filename: str = './odesystem', 
            template: str = None):
        """
        Produce ODE system, load a template file and write an executable python script.
        """

        if concvect and len(concvect) != len(sorted_vars):
            raise CRNSimulatorError('Concentrations cannot be mapped to species!')

        V, M, J, R = self.ode_system(sorted_vars = sorted_vars, 
                                     jacobian = jacobian, 
                                     rate_dict = rate_dict)

        return writeODElib(V, M, jacobian = J, rdict = R, concvect = concvect,
                           odename = odename, filename = filename, template = None)

    def ode_system(self, 
            sorted_vars: List[str] = None, 
            jacobian: bool = False, 
            rate_dict: bool = False) -> Tuple[List[str], 
                                              sM, 
                                              Union[sM, None], 
                                              Union[Dict[str, str], None]]:

        odict, R = self.get_odes(rate_dict = rate_dict)

        if sorted_vars:
            sorted_vars = list(map(Symbol, sorted_vars))
            assert len(sorted_vars) == len(list(odict.keys()))
        else:
            sorted_vars = sorted(list(odict.keys()), key=lambda x: str(x))

        # Symbol namespace dictionary, translates every variable name into a Symbol,
        #   even awkward names such as 'sin' or 'cos'
        ns = dict(zip(map(str, sorted_vars), sorted_vars))

        M = []
        for dx in sorted_vars:
            sfunc = sympify(
                ' + '.join(['*'.join(map(str, xp)) for xp in odict[dx]]), locals=ns)
            odict[dx] = sfunc
            M.append(sfunc)
        M = Matrix(M)

        if jacobian:
            # NOTE: The sympy version breaks regularly:
            # J = M.jacobian(sorted_vars)
            # ... so it is done per pedes:
            J = []
            for f in M:
                for x in sorted_vars:
                    J.append(f.diff(x))
            J = Matrix(J)
        else:
            J = None

        return sorted_vars, M, J, R 

    def get_odes(self, rate_dict: bool = False) -> Tuple[
            Dict[str, List[str]], Union[Dict[str, str], None]]:
        """Translate the reaction graph into ODEs.
        """
        rdict = dict()
        odes = dict()

        for rxn in self.reactions:
            if rate_dict:
                rate = 'k' + str(len(list(rdict.keys())))
                rdict[rate] = self._RG.nodes[rxn]['rate']
            else:
                rate = str(self._RG.nodes[rxn]['rate'])

            reactants = []
            for reac in self._RG.predecessors(rxn):
                for i in range(self._RG.number_of_edges(reac, rxn)):
                    reactants.append(Symbol(reac))

            products = []
            for prod in self._RG.successors(rxn):
                for i in range(self._RG.number_of_edges(rxn, prod)):
                    products.append(Symbol(prod))

            for x in reactants:
                if x in odes:
                    odes[x].append(['-' + rate] + reactants)
                else:
                    odes[x] = [['-' + rate] + reactants]

            for x in products:
                if x in odes:
                    odes[x].append([rate] + reactants)
                else:
                    odes[x] = [[rate] + reactants]

        return odes, rdict

    def add_reactions(self, crn: CRN) -> None:
        assert isinstance(crn, list)
        for rxn in crn:
            self.add_reaction(rxn)

    def add_reaction(self, rxn: RXN) -> None:
        assert len(rxn) == 3
        n = ReactionNode()

        if isinstance(rxn[2], list):
            # TODO: for now it is ok, but maybe we should enforce a consistent format here.
            #
            # sometimes the format is [[react],[prod], [k]],
            # sometimes it is [[react],[prod], k]
            #
            # print(DeprecationWarning('Using deprecated format for irreversible reactions.'))
            rxn[2] = rxn[2][0]

        self._RG.add_node(n, rate = rxn[2])
        for r in rxn[0]:
            assert isinstance(r, str)
            self._RG.add_edge(r, n)
        for p in rxn[1]:
            assert isinstance(p, str)
            self._RG.add_edge(n, p)

        return



#
# Parser for formal chemical reaction networks.
#
# Written by Stefan Badelt (badelt@caltech.edu).
#
# This file was adapted from the python package: `Nuskell`
# Developed at the *DNA and Natural Algorithms Group*, Caltech.
# originally written by Seung Woo Shin (seungwoo.theory@gmail.com).
#
# Use at your own risk.
#
#

from pyparsing import (Word, Literal, Group, Suppress, Combine, Optional,
                       alphas, nums, alphanums, delimitedList, StringStart, StringEnd, LineEnd,
                       ZeroOrMore, OneOrMore, pythonStyleComment, ParseElementEnhance)

from pyparsing import ParseException

class CRNParseError(Exception):
    pass

def crn_document_setup(modular=False):
    """Parse a formal chemical reaction network.

    Args:
      modular <optional:bool>: Adds an additional nesting for modules within a
        CRN. Use one line per module (';' separates reactions).

    Format:
      # A list of reactions, optionally with reaction rates:
      # <- this is a comment!
      B + B -> C    # [k = 1]
      C + A <=> D   # [kf = 1, kr = 1]
      <=> A  [kf = 15, kr = 6]

      # Note that you can write multiple reactions in one line:
      A + 2C -> E [k = 13.78]; E + F <=> 2A  [kf = 13, kr = 14]

    Returns:

    """
    # NOTE: If you want to add support for multiple modules per line, you can use
    # the '|' character.

    W = Word
    G = Group
    S = Suppress
    O = Optional
    C = Combine
    L = Literal

    def T(x, tag):
        """ Return a *Tag* to distinguish (ir)reversible reactions """
        def TPA(tag):
            return lambda s, l, t: [tag] + t.asList()
        return x.setParseAction(TPA(tag))

    crn_DWC = "".join(
        [x for x in ParseElementEnhance.DEFAULT_WHITE_CHARS if x != "\n"])
    ParseElementEnhance.setDefaultWhitespaceChars(crn_DWC)

    identifier = W(alphas, alphanums + "_")

    multiplier = W(nums)
    species = G(O(multiplier) + identifier)

    number = W(nums, nums)
    num_flt = C(number + O(L('.') + number))
    num_sci = C(number + O(L('.') + number) + L('e') + O(L('-') | L('+')) + W(nums))
    gorf = num_sci | num_flt

    # Make specification of forward, backward, reverse more flexible
    kf = S('kf') | S('fw')
    kr = S('kr') | S('bw') | S('rv')

    k = G(S('[') + O(S('k') + S('=')) + gorf + S(']'))
    rev_k = G(S('[') + kf + S('=') + gorf + S(',') + kr + S('=') + gorf + S(']')) | G(S('[') + gorf + S(',') + gorf + S(']'))

    concentration = T(species + S('@') + G(L("initial") | L("i") | L("constant") | L("c")) + G(gorf), 'concentration')

    reaction = T(G(O(delimitedList(species, "+"))) +
                 S("->") +
                 G(O(delimitedList(species, "+"))) + O(k), 'irreversible')

    rev_reaction = T(G(O(delimitedList(species, "+"))) +
                     S("<=>") +
                     G(O(delimitedList(species, "+"))) + O(rev_k), 'reversible')

    expr = G(reaction | rev_reaction | concentration)

    if modular:
        module = G(expr + ZeroOrMore(S(";") + expr))
    else:
        module = expr + ZeroOrMore(S(";") + expr)

    crn = OneOrMore(module + ZeroOrMore(S(LineEnd())))

    document = StringStart() + ZeroOrMore(S(LineEnd())) + crn + StringEnd()
    document.ignore(pythonStyleComment)
    return document

def post_process(crn):
    """Process a parsed CRN.

    Does:
     - remove 'reversible' and 'irreversible' tags.
        (this information is stored implicitly with one or two rates).
     - Translates species multipliers to multisets
     - Extracts the set of all species in the CRN.
     - Adds 'None' as reaction rate if no rate was specified.

    Does not:
     - Check if reactions are specified twice
     - Check stochiometry

    Returns:
        new: the CRN in format [[r],[p],k]
        species: dictionary species[name]=('initial', float)

    """
    def remove_multipliers(species):
        flat = []
        for s in species:
            if len(s) == 1:
                flat.append(s[0])
            elif len(s) == 2:
                ss = [s[1]] * int(s[0])
                flat.extend(ss)
        return flat

    new = []
    species = dict()
    for line in crn:
        if line[0] == 'concentration':
            spe = line[1][0]
            ini = 'initial' if line[2][0][0] == 'i' else 'constant'
            num = line[3][0]
            species[spe] = (ini, float(num))
            continue
        elif len(line) == 3:
            # No rate specified
            t, r, p = line
            r = remove_multipliers(r)
            p = remove_multipliers(p)
            if t == 'reversible':
                new.append([r, p, [None, None]])
            elif t == 'irreversible':
                new.append([r, p, [None]])
            else:
                raise CRNParseError('Wrong CRN format!')
        elif len(line) == 4:
            t, r, p, k = line
            r = remove_multipliers(r)
            p = remove_multipliers(p)
            if t == 'reversible':
                assert len(k) == 2
                new.append([r, p, k])
            elif t == 'irreversible':
                assert len(k) == 1
                new.append([r, p, k])
            else:
                raise CRNParseError('Wrong CRN format!')
        else:
            raise CRNParseError('Wrong CRN format!')
        for s in r + p:
            if s not in species:
                species[s] = ('initial', 0)

        #species = species.union(r).union(p)
        #print species
    return new, species

def parse_crn_file(filename, process=True):
    """Parses a CRN from a file.

    Args:
      filename (<str>): The name of the file for parsing.
      process (optional <bool>): Process the format of the returned CRN.

    Returns:
      crn (<lol>): List of list representation of a CRN
      species (<set()>): A set of all involved species (only when process=True)

    """
    crn_document = crn_document_setup()
    if process:
        return post_process(crn_document.parseFile(
            filename, parseAll=True).asList())
    else:
        return crn_document.parseFile(filename, parseAll=True).asList()

def parse_crn_string(data, process=True):
    """Parses a CRN from a string.

    Args:
      filename (<str>): The name of the file for parsing.
      process (optional <bool>): Process the format of the returned CRN.

    Returns:
      crn (<lol>): List of list representation of a CRN
      species (<set()>): A set of all involved species (only when process=True)

    """
    crn_document = crn_document_setup()
    if process:
        return post_process(crn_document.parseString(data).asList())
    else:
        return crn_document.parseString(data).asList()




#!/usr/bin/env python

from setuptools import setup, find_packages

LONG_DESCRIPTION = """
Simulate chemical recation networks (CRNs) using ordinary differential
equations (ODEs).
"""

setup(
    name = 'crnsimulator',
    version = "0.7.1",
    description = 'Simulate CRNs using ODEs.',
    long_description = LONG_DESCRIPTION,
    author = 'Stefan Badelt',
    author_email = 'badelt@caltech.edu',
    url = 'https://github.com/bad-ants-fleet/crnsimulator',
    download_url = 'https://github.com/bad-ants-fleet/crnsimulator/archive/v0.7.1.tar.gz',
    license = 'MIT',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.7',
        ],
    install_requires = [
        'scipy>=0.16.1',
        'sympy>=0.7.6.1',
        'pyparsing',
        'numpy',
        'matplotlib',
        'seaborn',
        'networkx>=2.2'],
    packages = ['crnsimulator'],
    scripts = ['scripts/crnsimulator'],
    test_suite='tests'
)




#!/usr/bin/env python3
#
# crnsimulator.odelib_template: A template for a compiled ODE solver.
# ... or an autogenerated script from the *crnsimulator* Python package
#
# Written by Stefan Badelt (badelt@caltech.edu).
#
# Use at your own risk.
#
#

#
# If this file is executable, it contains a system of hardcoded ODEs together
# with some default parameters. It is recommended to edit the source directly
# at "crnsimulator.odelib_template" or provide your own template file.
#
# call with: python #<&>FILENAME<&># --help
#

import argparse
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import odeint

import seaborn as sns
sns.set(style="darkgrid", font_scale=1, rc={"lines.linewidth": 2.0})


def ode_plotter(name, t, ny, svars, log = False, labels = None,
        xlim = None, ylim = None, plim = None):
    """ Plots the ODE trajectories.

    Args:
      name (str): Name of the outputfile including extension (e.g. *.pdf)
      t (list[flt]) : time units plotted on the x-axis.
      ny (list[list[flt]]) : a list of trajectories plotted on the y-axis.
      svars (list[str]): A list of names for every trajectory in ny
      log (bool,optional): Plot data on a logarithmic time scale
      labels (set(),optional): Define species that appear labelled in the plot
      xlim ((float,float), optional): matplotlib xlim.
      ylim ((float,float), optional): matplotlib ylim.
      plim (float, optional): Minimal occupancy to plot a trajectory. Defaults to None.

    Prints:
      A file containing the plot (Format *.pdf, *.png, etc.)

    Returns:
      [str]: Name of the file containing the plot
    """
    fig, ax = plt.subplots(1, 1, figsize=(8, 4.5))

    # b : blue.
    # g : green.
    # r : red.
    # c : cyan.
    # m : magenta.
    # y : yellow.
    # k : black.
    mycolors = ['blue', 
                'red', 
                'green', 
                'orange', 
                'maroon', 
                'springgreen', 
                'cyan', 
                'magenta', 
                'yellow']
    mycolors += list('kkkkkkkkkkk')

    if labels:
        i = 0
        for e, y in enumerate(ny):
            if svars[e] in labels:
                ax.plot(t, y, '-', label=svars[e], color=mycolors[i])
                i = i + 1 if i < len(mycolors) - 1 else 0
            else:
                ax.plot(t, y, '--', lw=0.1, color='gray', zorder=1)
    else:
        for e, y in enumerate(ny):
            if plim is None or max(y) > plim:
                ax.plot(t, y, '-', label=svars[e])
            else:
                ax.plot(t, y, '--', lw=0.1, color='gray', zorder=1)

    plt.title(name)
    if xlim:
        plt.xlim(xlim)
    # plt.xticks(np.arange(0, 61, step=20))

    if ylim:
        plt.ylim(ylim)
    # plt.yticks(np.arange(0, 51, step=10))

    ax.set_xlabel('Time', fontsize=16)
    ax.set_ylabel('Concentration', fontsize=16)
    if log:
        ax.set_xscale('log')
    else:
        ax.set_xscale('linear')

    plt.legend(loc='upper right')
    fig.tight_layout()
    plt.savefig(name)
    plt.close()
    return name


rates = {
    #<&>RATES<&>#
}

#<&>ODECALL<&>#

#<&>JACOBIAN<&>#


def add_integrator_args(parser):
    """ODE integration aruments."""
    solver = parser.add_argument_group('odeint parameters')
    plotter = parser.add_argument_group('plotting parameters')

    plotter.add_argument("--list", action='store_true',
            help="Print all species and exit.")

    # required: simulation time and output settings
    solver.add_argument("--t0", type=float, default=0, metavar='<flt>',
            help="First time point of the time-course.")
    solver.add_argument("--t8", type=float, default=100000, metavar='<flt>',
            help="End point of simulation time.")
    plotter.add_argument("--t-lin", type=int, default=50000, metavar='<int>',
            help="Returns --t-lin evenly spaced numbers on a linear scale from --t0 to --t8.")
    plotter.add_argument("--t-log", type=int, default=None, metavar='<int>',
            help="Returns --t-log evenly spaced numbers on a logarithmic scale from --t0 to --t8.")

    # required: initial concentration vector
    solver.add_argument("--p0", nargs='+', metavar='<int/str>=<flt>',
            help="""Vector of initial species concentrations. 
            E.g. \"--p0 1=0.5 3=0.7\" stands for 1st species at a concentration of 0.5 
            and 3rd species at a concentration of 0.7. You may chose to address species
            directly by name, e.g.: --p0 C=0.5.""")

    # optional: choose output formats
    plotter.add_argument("--nxy", action='store_true',
            help="Print time course to STDOUT in nxy format.")
    plotter.add_argument("--header", action='store_true',
            help="Print header for trajectories.")
    plotter.add_argument("--pyplot", default='', metavar='<str>',
            help="Specify a filename to plot the ODE simulation.")
    plotter.add_argument("--pyplot-labels", nargs='+', default=[], metavar='<str>+',
            help="Specify the species which should appear in the pyplot legend.")
    plotter.add_argument("--pyplot-xlim", nargs=2, type=float, default=None, metavar='<flt>',
            help="Specify the limits of the x-axis.")
    plotter.add_argument("--pyplot-ylim", nargs=2, type=float, default=None, metavar='<flt>',
            help="Specify the limits of the y-axis.")

    # advanced: scipy.integrate.odeint parameters
    solver.add_argument("-a", "--atol", type=float, default=None, metavar='<flt>',
            help="Specify absolute tolerance for the solver.")
    solver.add_argument("-r", "--rtol", type=float, default=None, metavar='<flt>',
            help="Specify relative tolerance for the solver.")
    solver.add_argument("--mxstep", type=int, default=0, metavar='<int>',
            help="Maximum number of steps allowed for each integration point in t.")

    return


def integrate(args):
    """Main interface to solve the ODE-system.

    Args:
      args (:obj:`argparse.ArgumentParser()`): An argparse object containing all of
        the arguments of :obj:`crnsimulator.add_integrator_args()`.

    Prints:
      - verbose information
      - plot files
      - time-course

    Returns:
      Nothing
    """

    #<&>SORTEDVARS<&>#

    p0 = [0] * len(svars)
    #<&>DEFAULTCONCENTRATIONS<&>#

    if args.list:
        for e, v in enumerate(svars, 1):
            print('# {} {}'.format(e, v))
        raise SystemExit('# Specify a vector of initial concentrations: ' +
                'e.g. --p0 1=0.1 2=0.005 3=1e-6 (see --help)')

    if not args.nxy and not args.pyplot:
        print('# Use --pyplot and/or --nxy to plot your results.')

    if not args.t8:
        raise ValueError(
            'Specify a valid end-time for the simulation: --t8 <flt>')

    if args.t_log:
        if args.t0 == 0:
            raise ValueError('--t0 cannot be 0 when using log-scale!')
        time = np.logspace(np.log10(args.t0), np.log10(args.t8), num=args.t_log)
    elif args.t_lin:
        time = np.linspace(args.t0, args.t8, num=args.t_lin)
    else:
        raise ValueError('Please specify either --t-lin or --t-log. (see --help)')

    if not args.p0:
        print('# Initial concentrations:', list(zip(svars, p0)))
        if sum(p0) == 0:
            for e, v in enumerate(svars, 1):
                print('# {} {}'.format(e, v))
            raise SystemExit('# -- Must specify a vector of initial concentrations: ' +
                             'e.g. --p0 1=0.1 2=0.005 3=1e-6 (see --help)')
    else:
        for term in args.p0:
            p, o = term.split('=')
            try:
                pi = svars.index(p)
            except ValueError as e:
                pi = int(p) - 1
            finally:
                p0[pi] = float(o)
        print('# Initial concentrations:', list(zip(svars, p0)))
        p0 = np.array(p0)

    # TODO: It would be nice if it is possible to read alternative rates from a file instead.
    # None triggers the default-rates that are hard-coded in the (this)
    # library file.
    rates = None

    ny = odeint(#<&>ODENAME<&>#,
        p0, time, (rates, ), #<&>JCALL<&>#,
        atol=args.atol, rtol=args.rtol, mxstep=args.mxstep).T

    if args.nxy:
        if args.header:
            print(' '.join(['{:15s}'.format(x) for x in ['time'] + svars]))
        for i in zip(time, *ny):
            print(' '.join(map("{:.9e}".format, i)))

    if args.pyplot:
        plotfile = ode_plotter(args.pyplot, time, ny, svars,
                               log=True if args.t_log else False,
                               labels=set(args.pyplot_labels),
                               xlim = args.pyplot_xlim,
                               ylim = args.pyplot_ylim)
        print('# Printed file:', plotfile)

    return zip(time, *ny)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    add_integrator_args(parser)

    args = parser.parse_args()

    integrate(args)
